import uuid
import json
import pika
from django.utils import timezone
import logging
import importlib



class VirtualHost(object):
    """
    A RabbitMQ virtual host. Takes the following parameters:

    :param str host: The url to the host, e.g. 192.168.0.1 or *localhost*
    :param str name: The virtual host name, eg "myvirtualhost"
    :param int port: defaults to 5672
    :param str username: RabbitMQ username
    :param str password: RabbitMQ password
    :param bool secure: Whether or not to use SSL. Defaults to False

    """
    def __init__(self, url=None, host='localhost', name='/', port=5672, username='guest', password='guest',
                 secure=False):

        self.secure = secure
        if not url:
            self.host = host
            self.name = name
            self.port = port
            self.username = username
            self.password = password

        else:
            try:
                url = url.replace('amqp://', '')
                if '@' in url:
                    # user:pass@host:port/vhost
                    credentials, url = url.split('@')
                    self.username, self.password = credentials.split(':')

                else:
                    # host:port/vhost
                    self.username = username
                    self.password = password

                try:
                    url, self.name = url.split('/')
                except ValueError:
                    url = url.split('/')[0]
                    self.name = name

                self.host, _port = url.split(':')
                self.port = int(_port)

            except Exception as err:
                raise Exception('Unable to parse the RabbitMQ server. Please check your configuration: %s' % err)

    def __str__(self):
        """
        Returns the broker url
        """
        return 'amqp://%s:%s@%s:%s/%s' % (self.username, self.password, self.host, self.port, self.name)

    @property
    def blocking_connection(self):
        """
        Connect to the VHOST

        :return: a pika.BlockingConnection object
        """
        credentials = pika.PlainCredentials(username=self.username, password=self.password)
        params = pika.ConnectionParameters(host=self.host, port=self.port, virtual_host=self.name,
                                           credentials=credentials, connection_attempts=10, ssl=self.secure,
                                           heartbeat=1200)
        return pika.BlockingConnection(parameters=params)


class BaseMessageSerializer(object):
    """
    A class that defines how to convert a RabbitMQ message into an executable python function from your Django project,
    and back again

    :param Message message: the RabbitMQ message

    """
    content_type = 'application/json'
    type_header, message_type = (None,) * 2

    def get_task(self, properties, body):
        """
        Identifies the python function to be executed from the content of the RabbitMQ message. By default, Carrot
        returns the value of the self.type_header header in the properties.

        Once this string has been found, carrot uses importlib to return a callable python function.

        :param properties: the message properties
        :param body: the message body. This parameter is not used by default, but is provided so that this method can
                     be extended to identify messages based on the content from the body, instead of the properties, in
                     custom consumers

        :return: a callable python function

        """
        mod = '.'.join(properties.headers[self.type_header].split('.')[:-1])
        task = properties.headers[self.type_header].split('.')[-1]
        module = importlib.import_module(mod)
        func = getattr(module, task)
        return func

    def __init__(self, message=None):
        self.message = message

    def publish_kwargs(self):
        """
        Returns a dictionary of keyword arguments to be passed to channel.basic_publish. In this implementation, the
        exchange, routing key and message body are returned

        :rtype: dict

        """
        exchange = self.message.exchange or ''
        routing_key = self.message.routing_key or 'default'
        body = self.body()
        return dict(exchange=exchange, routing_key=routing_key, body=body, mandatory=True)

    def body(self):
        """
        Returns the content to be added to the RabbitMQ message body

        By default, this implementation returns a simple dict in the following format:

        .. code-block:: python

            {
                'args': ('tuple', 'of', 'positional', 'arguments'),
                'kwargs': {
                    'keyword1': 'value',
                    'keyword2': True
                }
            }

        :rtype: dict

        """
        args = self.message.task_args
        kwargs = self.message.task_kwargs

        data = {}

        if args:
            data['args'] = args

        if kwargs:
            data['kwargs'] = kwargs

        return json.dumps(data)

    def properties(self):
        """
        Returns a dict from which a :class:`pika.BasicProperties` object can be created

        In this implementation, the following is returned:
        - headers
        - content type
        - priority
        - message id
        - message type

        :rtype: dict

        """
        headers = {self.type_header: self.message.task}
        content_type = self.content_type
        priority = self.message.priority
        message_id = str(self.message.uuid)
        type = self.message_type

        return dict(headers=headers, content_type=content_type, priority=priority, message_id=message_id, type=type)

    def publish(self, connection, channel):
        kwargs = self.publish_kwargs()
        kwargs['properties'] = pika.BasicProperties(**self.properties())
        channel.basic_publish(**kwargs)
        connection.close()

    def serialize_arguments(self, body):
        """
        Extracts positional and keyword arguments to be sent to a function from the message body

        :param str body: the message body as a JSON string
        :return: a tuple of the positional and keyword arguments

        """
        content = json.loads(body)
        args = content.get('args', ())
        kwargs = content.get('kwargs', {})
        return args, kwargs


class DefaultMessageSerializer(BaseMessageSerializer):
    type_header = 'type'
    message_type = 'django-carrot message'


class Message(object):
    """
    A message to publish to RabbitMQ. Takes the following parameters:

    :param str task: the path to the task to execute, e.g. *myapp.mymodule.myfunction*
    :param VirtualHost virtual_host: The host containing the queue to publish the message to
    :param str queue: the queue name. Will be set to *default* if not provided
    :param str routing_key: the routing key. Defaults to the queue name
    :param str exchange: the RabbitMQ exchange name. Can be blank (i.e. a direct exchange)
    :param tuple task_args: positional arguments to be passed to the task
    :param dict task_kwargs: keyword arguments to be sent to the task
    :param int priority: a priority between 0 and 255, where 255 is the highest priority

    .. note::
        Your RabbitMQ queue must support message priority for the *priority* parameter to have any affect. You need to
        define the x-max-priority header when creating your RabbitMQ queue to do this. See
        `Priority Queue Support <https://www.rabbitmq.com/priority.html>`_ for more details. Carrot applies a maximum
        priority of **255** by default to all queues it creates automatically.

    .. warning::
        You should not attempt to create instances of this object yourself. You should use the
        :func:`carrot.utilities.create_msg` function instead

    """
    def __init__(self, task, virtual_host=None, queue='default', routing_key=None, exchange='', priority=0,
                 task_args=(), task_kwargs=None):
        if not task_kwargs or task_kwargs in ['{}', '"{}"']:
            task_kwargs = {}

        if not routing_key:
            routing_key = queue

        if not virtual_host:
            from carrot.utilities import get_host_from_name
            virtual_host = get_host_from_name(queue)

        assert isinstance(virtual_host, VirtualHost)

        self.uuid = str(uuid.uuid4())

        self.virtual_host = virtual_host
        self.exchange = exchange
        self.queue = queue
        self.routing_key = routing_key
        self.priority = priority

        self.task = task
        self.task_args = task_args
        self.task_kwargs = task_kwargs

        self.formatter = DefaultMessageSerializer(self)

    @property
    def connection_channel(self):
        """
        Gets or creates the queue, and returns a tuple containing the object's VirtualHost's blocking connection,
        and its channel
        """
        connection = self.virtual_host.blocking_connection
        channel = connection.channel()

        return connection, channel

    def publish(self, pika_log_level=logging.ERROR):
        """
        Publishes the message to RabbitMQ queue and creates a MessageLog object so the progress of the task can be
        tracked in the Django project's database

        :param pika_log_level: the pika log level to set. defaults to logging.ERROR
        :type pika_log_level: logging level

        """
        logging.getLogger("pika").setLevel(pika_log_level)
        connection, channel = self.connection_channel
        from carrot.models import MessageLog

        try:
            keyword_arguments = json.dumps(json.loads(self.task_kwargs or '{}') or {})
        except TypeError:
            keyword_arguments = json.dumps(self.task_kwargs)

        log = MessageLog.objects.create(
            status='PUBLISHED',
            queue=self.queue,
            exchange=self.exchange or '',
            routing_key=self.routing_key or self.queue,
            uuid=str(self.uuid),
            priority=self.priority,
            task_args=self.task_args,
            content=keyword_arguments,
            task=self.task,
            publish_time=timezone.now(),
        )

        self.formatter.publish(connection, channel)
        return log




from django.conf.urls import url
from carrot.views import MessageList
from carrot.utilities import decorate_class_view, decorate_function_view
from django.conf import settings
from carrot.api import (
    published_message_log_viewset, failed_message_log_viewset, completed_message_log_viewset, scheduled_task_viewset,
    detail_message_log_viewset, scheduled_task_detail
)


decorators = settings.CARROT.get('monitor_authentication', [])


def _(v, **kwargs):
    return decorate_class_view(v, decorators).as_view(**kwargs)


def _f(v):
    return decorate_function_view(v, decorators)


urlpatterns = [
    url(r'^$', _(MessageList), name='carrot-monitor'),
    url(r'^api/message-logs/published/$', _f(published_message_log_viewset), name='published-messagelog'),
    url(r'^api/message-logs/failed/$', _f(failed_message_log_viewset)),
    url(r'^api/message-logs/completed/$', _f(completed_message_log_viewset)),
    url(r'^api/message-logs/(?P<pk>[0-9]+)/$', _f(detail_message_log_viewset)),
    url(r'^api/scheduled-tasks/$', _f(scheduled_task_viewset)),
    url(r'^api/scheduled-tasks/(?P<pk>[0-9]+)/$', _f(scheduled_task_detail)),
]


from django import template
from django.utils.html import mark_safe
from django.conf import settings
import importlib
from inspect import getmembers, isfunction
from carrot.models import ScheduledTask
from django.template.response import SimpleTemplateResponse


register = template.Library()


def get_content(template, context=None):
    if context is None:
        context = {}
    task_temp = SimpleTemplateResponse('carrot/%s.html' % template, context=context)
    content = task_temp.rendered_content.replace('\n', '')
    return mark_safe(content)


@register.inclusion_tag('carrot/carrot.js')
def vue():
    task_content = get_content('task_detail')
    field_errors_content = get_content('field-errors')
    scheduled_content = get_content('scheduled_task_detail')
    paginator_content = get_content('paginator')
    search_bar_content = get_content('search_bar')

    modules = settings.CARROT.get('task_modules', [])
    task_choices = []
    interval_choices = [c[0] for c in ScheduledTask.INTERVAL_CHOICES]
    for module in modules:
        try:
            mod = importlib.import_module(module)
            functions = [o[0] for o in getmembers(mod) if isfunction(o[1]) and not o[0] == 'task']

            for function in functions:
                f = '%s.%s' % (module, function)
                task_choices.append(f)
        except (ImportError, AttributeError):
            pass

    return {
        'task_detail_template': task_content,
        'scheduled_task_template': scheduled_content,
        'task_options': mark_safe(task_choices),
        'interval_options': mark_safe(interval_choices),
        'field_errors': field_errors_content,
        'paginator_template': paginator_content,
        'search_bar_template': search_bar_content,
    }




"""
This module provides a backend API for creating Consumers and Consumer Sets

"""

import json
from carrot.models import MessageLog
import traceback
import threading
import logging
from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone
import importlib
from carrot.objects import DefaultMessageSerializer
from django.db.utils import OperationalError
import pika
import sys
from django.conf import settings


LOGGING_FORMAT = '%(threadName)-10s %(asctime)-10s %(levelname)s:: %(message)s'


class Consumer(threading.Thread):
    """
    An individual Consumer object. This class is run on a detached thread and watches a specific RabbitMQ queue for
    messages, and consumes them when they appear. Multiple Consumers can be linked to the same queue using a
    :class:`.ConsumerSet` object.
    """

    serializer = DefaultMessageSerializer()
    reconnect_timeout = 5
    task_log = []
    queue_arguments = {}
    exchange_arguments = {}
    active_message_log = None
    remaining_save_attempts = 10

    def __init__(self, host, queue, logger, name, durable=True, queue_arguments=None, exchange_arguments=None):
        """
        :param host: the host the queue to consume from is attached to
        :type host: :class:`carrot.objects.VirtualHost`
        :param str queue: the queue to consume from
        :param logger: the logging object
        :param name: the name of the consumer

        """
        super(Consumer, self).__init__()

        if queue_arguments is None:
            queue_arguments = {}

        if not exchange_arguments:
            exchange_arguments = {}

        self.name = name
        self.logger = logger
        self.queue = queue
        self.exchange = queue

        self.connection = None
        self.channel = None
        self.shutdown_requested = False
        self._consumer_tag = None
        self._url = str(host)

        self.queue_arguments = queue_arguments
        self.exchange_arguments = exchange_arguments
        self.durable = durable

    def fail(self, log, err):
        """
        This function is called if there is any kind of error with the `.consume()` function

        :param MessageLog log: the associated MessageLog object
        :param str err: the exception

        The exception message is logged, and the MessageLog is updated with the result

        """
        self.logger.error('Task %s failed due to the following exception: %s' % (log.task, err))

        if self.task_log:
            log.log = '\n'.join(self.task_log)

        log.status = 'FAILED'
        log.failure_time = timezone.now()
        log.exception = err
        log.traceback = traceback.format_exc()
        log.save()

    def get_task_type(self, properties, body):
        """
        Identifies the task type, by looking up the attribute **self.task_type** in the message properties

        :param properties: the message properties
        :param body: the message body. Not used by default, but provided so that the method can be extended if necessary

        :return: The task type as a string, e.g. *myapp.mymodule.mytask*

        """
        return properties[self.serializer.type_header]

    def get_message_log(self, properties, body):
        """
        Finds and returns the :class:`carrot.models.MessageLog` object associated with a RabbitMQ message

        By default, carrot finds this retrieving the MessageLog UUID from the RabbitMQ message properties.message_id
        attribute.

        This method can be extended by custom consumers. For example, if you are attempting to consume from a RabbitMQ
        queue containing messages that do not come from your Carrot instance, you may want to extend this method to
        create, instead of get, a MessageLog object

        :param properties: the message properties
        :param body: the message body. This is not used by default, but is included so that the function can be
                     extended in custom consumers.

        :rtype: class:`carrot.models.MessageLog` or None

        In order to avoid different consumers picking up the same message, MessageLogs are only
        .. note::
            This method does not use self.get_task_type as the intention is to get the MessageLog object before the
            **consume** method tries to do anything else. This means that if any later part of the process fails,
            the traceback and exception information can be stored to the MessageLog object for easier debugging.

        .. warning::
            If this method fails to find a matching MessageLog object, then the RabbitMQ message will be rejected.
            Depending on the configuration of your RabbitMQ queue, this may cause a loss of data. If you are
            implementing a custom consumer, then you should use
            `dead letter exchange <http://www.rabbitmq.com/dlx.html>`_ to preserve your message content

        """
        try:
            log = MessageLog.objects.get(uuid=properties.message_id)
        except ObjectDoesNotExist:
            return

        if log.status == 'PUBLISHED':
            return log

    def connect(self):
        """
        Connects to the broker

        :rtype: pika.SelectConnection

        """
        self.logger.info('Connecting to %s', self._url)
        return pika.SelectConnection(pika.URLParameters(self._url), self.on_connection_open, stop_ioloop_on_close=False)

    def on_connection_open(self, connection):
        """
        Callback that gets called when the connection is opened. Adds callback in case of a closed connection, and
        establishes the connection channel

        :param connection: Sent by default by pika but not used by carrot
        :type connection: pika.SelectConnection

        """
        self.logger.info('Connection opened')
        self.connection.add_on_close_callback(self.on_connection_closed)
        self.connection.channel(on_open_callback=self.on_channel_open)

    def on_connection_closed(self, *args):
        """
        Callback that gets called when the connection is closed. Checks for the self.shutdown_requested parameter first,
        which is used to idenfity whether the shutdown has been requested by the user or not. If not, carrot attempts to
        reconnect

        All arguments sent to this callback come from Pika but are not required by Carrot
        """

        self.channel = None
        if self.shutdown_requested:
            self.logger.warning('Connection closed')
            self.connection.ioloop.stop()
            self.logger.warning('Connection IO loop stopped')
        else:
            self.logger.warning('Connection closed unexpectedly. Trying again in %i seconds' % self.reconnect_timeout)
            self.connection.add_timeout(self.reconnect_timeout, self.reconnect)

    def reconnect(self):
        """
        Reconnect to the broker in case of accidental disconnection
        """
        self.connection.ioloop.stop()

        if not self.shutdown_requested:
            self.connection = self.connect()
            self.connection.ioloop.start()

    def on_channel_open(self, channel):
        """
        This function is invoked when the channel is established. It adds a callback in case of channel closure, and
        establishes the exchange

        :param pika.channel.Channel channel: The channel object

        """
        self.logger.info('Channel opened')
        self.channel = channel
        self.channel.add_on_close_callback(self.on_channel_closed)
        self.channel.exchange_declare(self.on_exchange_declare, self.exchange, **self.exchange_arguments)

    def on_channel_closed(self, channel, reply_code, reply_text):
        """
        Called when the channel is closed. Raises a warning and closes the connection

        Parameters are provided by Pika but not required by Carrot
        """
        if not self.shutdown_requested:
            self.logger.warning('Consumer %s not running: %s' % (self.name, reply_text))
            if self.active_message_log:
                self.active_message_log.requeue()
        else:
            self.logger.warning('Channel closed by client. Closing the connection')

        self.connection.close()

    def on_exchange_declare(self, *args):
        """
        Invoked when the exchange has been successfully established

        Parameters are provided by Pika but not required by Carrot
        """
        self.logger.info('Exchange declared')
        self.channel.queue_declare(self.on_queue_declare, self.queue, durable=self.durable,
                                   arguments=self.queue_arguments)

    def on_queue_declare(self, *args):
        """
        Invoked when the queue has been successfully declared

        Parameters are provided by Pika but not required by Carrot
        """
        self.channel.queue_bind(self.on_bind, self.queue, self.exchange)

    def on_bind(self, *args):
        """
        Invoked when the queue has been successfully bound to the exchange

        Parameters are provided by Pika but not required by Carrot
        """
        self.logger.info('Queue bound')
        self.start_consuming()

    def start_consuming(self):
        """
        The main consumer process. Attaches a callback to be invoked whenever there is a new message added to the queue
        """
        self.logger.info('Starting consumer %s' % self.name)
        self.channel.add_on_cancel_callback(self.on_consumer_cancelled)
        self._consumer_tag = self.channel.basic_consume(self.on_message, self.queue)

    def on_consumer_cancelled(self, method_frame):
        """
        Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame

        """
        self.logger.warning('Consumer was cancelled remotely, shutting down: %r', method_frame)
        if self.channel:
            self.channel.close()

    def on_message(self, channel, method_frame, properties, body):
        """
        The process that takes a single message from RabbitMQ, converts it into a python executable and runs it,
        logging the output back to the assoicated :class:`carrot.models.MessageLog`

        :param channel: not used
        :type channel: pika.channel.Channel
        :param method_frame: contains the delivery tag
        :type method_frame: pika.Spec.Basic.Deliver
        :param properties: the message properties
        :type properties: pika.Spec.BasicProperties
        :param bytes body: The message body
        """
        log = self.get_message_log(properties, body)
        if log:
            self.active_message_log = log
            self.channel.basic_ack(method_frame.delivery_tag)
            log.status = 'IN_PROGRESS'
            log.save()
        else:
            self.logger.error('Unable to find a MessageLog matching the uuid %s. Ignoring this task' %
                              properties.message_id)
            self.channel.basic_nack(method_frame.delivery_tag, requeue=False)
            return

        try:
            task_type = self.get_task_type(properties.headers, body)
        except KeyError as err:
            self.fail(log, 'Unable to identify the task type because a key was not found in the message header: %s' %
                      err)
            return

        self.logger.info('Consuming task %s, ID=%s' % (task_type, properties.message_id))

        try:
            func = self.serializer.get_task(properties, body)
        except (ValueError, ImportError, AttributeError) as err:
            self.fail(log, err)
            return

        try:
            args, kwargs = self.serializer.serialize_arguments(body.decode())

        except Exception as err:
            self.fail(log, 'Unable to process the message due to an error collecting the task arguments: %s' % err)
            return

        start_msg = '{} {} INFO:: Starting task {}.{}'.format(self.name,
                                                              timezone.now().strftime("%Y-%m-%d %H:%M:%S,%f")[:-3],
                                                              func.__module__, func.__name__)
        self.logger.info(start_msg)
        self.task_log = [
            start_msg
        ]
        task = LoggingTask(func, self.logger, self.name, *args, **kwargs)

        try:
            output = task.run()

            self.task_log.append(task.get_logs())
            self.logger.info(task.get_logs())

            success = '{} {} INFO:: Task {} completed successfully with response {}'.format(self.name,
                                                                                            timezone.now().strftime(
                                                                                                "%Y-%m-%d %H:%M:%S,%f")[
                                                                                            :-3],
                                                                                            log.task, output)
            self.logger.info(success)
            self.task_log.append(success)

            log.status = 'COMPLETED'
            log.completion_time = timezone.now()

            if isinstance(output, dict):
                log.output = json.dumps(output)
            else:
                log.output = output

            log.log = '\n'.join(self.task_log)

            save_attempts = 0

            while True:
                save_attempts += 1

                try:
                    log.save()
                    self.active_message_log = None

                    break
                except OperationalError:
                    if save_attempts > 10:
                        raise OperationalError('Unable to access the database. This is probably because the number '
                                               'of carrot threads is too high. Either reduce the amount of '
                                               'scheduled tasks consumers, or increase the max number of '
                                               'connections supported by your database')
                    self.connection.sleep(10)

        except Exception as err:
            self.task_log.append(task.get_logs())
            self.fail(log, str(err))

    def stop_consuming(self):
        """
        Stops the consumer and cancels the channel
        """
        if self.channel:
            self.shutdown_requested = True
            self.logger.warning('Shutdown received. Cancelling the channel')
            self.channel.basic_cancel(self.on_cancel, self._consumer_tag)

    def on_cancel(self, *args):
        """
        Invoked when the channel cancel is completed.

        Parameters provided by Pika but not required by Carrot

        """
        self.logger.info('Closing the channel')
        self.channel.close()

    def run(self):
        """
        Process starts here
        """
        self.connection = self.connect()
        self.connection.ioloop.start()

    def stop(self):
        """
        Cleanly exit the Consumer
        """
        self.logger.info('Stopping')
        self.shutdown_requested = True
        if self.channel:
            self.stop_consuming()
            self.connection.ioloop.start()
            self.logger.info('Consumer closed')
        else:
            self.logger.warning('Not running!')

    def close_connection(self):
        """This method closes the connection to RabbitMQ."""
        self.logger.info('Closing connection')
        self.connection.close()


class ListHandler(logging.Handler):
    """
    A :class:`logging.Handler` that records each log entry to a python list object, provided that the entry is coming
    from the correct thread.

    Allows for task-specific logging

    """
    def __init__(self, thread_name, level):
        self.output = []
        self.thread_name = thread_name
        super(ListHandler, self).__init__(level)

    def emit(self, record):
        msg = self.format(record)
        if msg.startswith(self.thread_name):
            self.output.append(msg)

    def parse_output(self):
        return '\n'.join([str(line) for line in self.output])


class LoggingTask(object):
    """
    Turns a function into a class with :meth:`.run()` method, and attaches a :class:`ListHandler` logging handler
    """
    def __init__(self, task, logger, thread_name, *args, **kwargs):
        self.task = task
        self.args = args
        self.kwargs = kwargs

        self.logger = logger
        self.thread_name = thread_name
        self.out = []
        self.stream_handler = ListHandler(thread_name, self.logger.getEffectiveLevel())
        self.stream_handler.setLevel(self.logger.getEffectiveLevel())
        formatter = logging.Formatter(LOGGING_FORMAT)
        self.stream_handler.setFormatter(formatter)

        self.logger.addHandler(self.stream_handler)

        self._keep_alive = None

    def run(self):
        output = self.task(*self.args, **self.kwargs)
        return output

    def get_logs(self):
        try:
            self.logger.removeHandler(self.stream_handler)
            return self.stream_handler.parse_output()
        except:
            return


class ConsumerSet(object):
    """
    Creates and starts a number of :class:`.Consumer` objects. All consumers must belong to the same queue

    :param host: The virtual host where the queue belongs
    :param queue: The queue name
    :param concurrency: the number of consumers to create. Defaults to 1
    :param name: the name to assign to the individual consumers. Will be rendered as *Consumer-1, Consumer-2,* etc.
    :param logfile: the path to the log file. Defaults to carrot.log
    :param loglevel: the logging level. Defaults to logging.DEBUG

    """
    durable = True
    queue_arguments = {'x-max-priority': 255}
    exchange_arguments = {}

    @staticmethod
    def get_consumer_class(consumer_class):
        module = '.'.join(consumer_class.split('.')[:-1])
        _cls = consumer_class.split('.')[-1]
        mod = importlib.import_module(module)
        return getattr(mod, _cls)

    def __init__(self, host, queue, logger, concurrency=1, name='consumer', consumer_class='carrot.consumer.Consumer'):
        self.logger = logger
        self.host = host
        self.connection = host.blocking_connection
        self.channel = self.connection.channel()
        self.queue = queue

        self.concurrency = concurrency
        self.name = '%s-%s' % (self.queue, name)
        self.consumer_class = self.get_consumer_class(consumer_class)
        self.threads = []

        queue_settings = [q for q in settings.CARROT.get('queues', []) if q.get('name', None) == self.queue]
        if queue_settings:
            q_settings = queue_settings[0]
            if q_settings.get('durable', None):
                self.durable = q_settings['durable']

            if q_settings.get('queue_arguments', None):
                self.queue_arguments = q_settings['queue_arguments']

            if q_settings.get('exchange_arguments', None):
                self.exchange_arguments = q_settings['exchange_arguments']

    def stop_consuming(self):
        """
        Stops all running threads. Loops through the threads twice - firstly, to set the signal to **False** on all
        threads, secondly to wait for them all to finish

        If a single loop was used here, the latter threads could still consume new tasks while the parent process waited
        for the earlier threads to finish. The second loop allows for quicker consumer stoppage and stops all consumers
        from consuming new tasks from the moment the signal is received
        """
        for t in self.threads:
            t.stop()

        for t in self.threads:
            print('Closing consumer %s' % t)
            t.join()
            print('Closed consumer %s' % t)

    def start_consuming(self):
        """
        Creates a thread for each concurrency level, e.g. if concurrency is set to 5, 5 threads are created.

        A :class:`.Consumer` is attached to each thread and is started
        """
        for i in range(0, self.concurrency):
            consumer = self.consumer_class(host=self.host, queue=self.queue, logger=self.logger,
                                           name='%s-%i' % (self.name, i + 1),
                                           durable=self.durable, queue_arguments=self.queue_arguments,
                                           exchange_arguments=self.exchange_arguments)
            self.threads.append(consumer)
            consumer.start()


